@@ -24,6 +24,7 @@
 
 #include "heif.h"
 #include "error.h"
+#include "box.h" // only for color_profile, TODO: maybe move the color_profiles to its own header
 
 #include <vector>
 #include <memory>
@@ -33,6 +34,9 @@
 
 namespace heif {
 
+  int chroma_h_subsampling(heif_chroma c);
+  int chroma_v_subsampling(heif_chroma c);
+
 class HeifPixelImage : public std::enable_shared_from_this<HeifPixelImage>,
                        public ErrorBuffer
 {
@@ -42,7 +46,7 @@ class HeifPixelImage : public std::enable_shared_from_this<HeifPixelImage>,
 
   void create(int width,int height, heif_colorspace colorspace, heif_chroma chroma);
 
-  void add_plane(heif_channel channel, int width, int height, int bit_depth);
+  bool add_plane(heif_channel channel, int width, int height, int bit_depth);
 
   bool has_channel(heif_channel channel) const;
 
@@ -63,12 +67,14 @@ class HeifPixelImage : public std::enable_shared_from_this<HeifPixelImage>,
 
   std::set<enum heif_channel> get_channel_set() const;
 
+  int get_storage_bits_per_pixel(enum heif_channel channel) const;
+
   int get_bits_per_pixel(enum heif_channel channel) const;
 
   uint8_t* get_plane(enum heif_channel channel, int* out_stride);
   const uint8_t* get_plane(enum heif_channel channel, int* out_stride) const;
 
-  void copy_new_plane_from(const std::shared_ptr<HeifPixelImage> src_image,
+  void copy_new_plane_from(const std::shared_ptr<const HeifPixelImage>& src_image,
                            heif_channel src_channel,
                            heif_channel dst_channel);
   void fill_new_plane(heif_channel dst_channel, uint8_t value, int width, int height);
@@ -77,9 +83,6 @@ class HeifPixelImage : public std::enable_shared_from_this<HeifPixelImage>,
                                     heif_channel src_channel,
                                     heif_channel dst_channel);
 
-  std::shared_ptr<HeifPixelImage> convert_colorspace(heif_colorspace colorspace,
-                                                     heif_chroma chroma) const;
-
   Error rotate_ccw(int angle_degrees,
                    std::shared_ptr<HeifPixelImage>& out_img);
 
@@ -94,13 +97,18 @@ class HeifPixelImage : public std::enable_shared_from_this<HeifPixelImage>,
 
   Error scale_nearest_neighbor(std::shared_ptr<HeifPixelImage>& output, int width,int height) const;
 
+  void set_color_profile(std::shared_ptr<const color_profile> profile) { m_color_profile = profile; }
+
+  std::shared_ptr<const color_profile> get_color_profile() { return m_color_profile; }
+
  private:
   struct ImagePlane {
     int width;
     int height;
     int bit_depth;
 
-    std::vector<uint8_t> mem;
+    uint8_t* mem; // aligned memory start
+    std::vector<uint8_t> unaligned_mem;
     int stride;
   };
 
@@ -108,19 +116,15 @@ class HeifPixelImage : public std::enable_shared_from_this<HeifPixelImage>,
   int m_height = 0;
   heif_colorspace m_colorspace = heif_colorspace_undefined;
   heif_chroma m_chroma = heif_chroma_undefined;
+  std::shared_ptr<const color_profile> m_color_profile;
 
   std::map<heif_channel, ImagePlane> m_planes;
-
-  std::shared_ptr<HeifPixelImage> convert_YCbCr420_to_RGB() const;
-  std::shared_ptr<HeifPixelImage> convert_YCbCr420_to_RGB24() const;
-  std::shared_ptr<HeifPixelImage> convert_YCbCr420_to_RGB32() const;
-  std::shared_ptr<HeifPixelImage> convert_RGB_to_RGB24_32() const;
-  std::shared_ptr<HeifPixelImage> convert_mono_to_RGB(int bpp) const;
-  std::shared_ptr<HeifPixelImage> convert_mono_to_YCbCr420() const;
-  std::shared_ptr<HeifPixelImage> convert_RGB24_32_to_YCbCr420() const;
 };
 
 
+ std::shared_ptr<HeifPixelImage> convert_colorspace(const std::shared_ptr<HeifPixelImage>& input,
+                                                    heif_colorspace colorspace,
+                                                    heif_chroma chroma);
 }
 
 #endif
